/*
 * Copyright (c) 2005 DocJava, Inc. All Rights Reserved.
 *
 * Anne MacIlvain - modified to do the final.
 * 
 */
package math.fourierTransforms.r2;

import j2d.ImageUtils;
import j2d.ShortImageBean;
import math.complex.ComplexFloat1d;

import java.awt.*;

public class FFT2dComplex {
    ShortImageBean sib;
    ComplexFloat1d r;
    ComplexFloat1d g;
    ComplexFloat1d b;

    public static void main(String[] args) {
        final Image image1 = ImageUtils.getImage();
        final Image image2 = ImageUtils.getImage();

        FFT2dComplex f1 = new FFT2dComplex(image1);
        FFT2dComplex f2 = new FFT2dComplex(image2);

        f1.normalize();
        f2.normalize();
        //make f2 unitary.
        f2.scale(1f / 255f);

        ImageUtils.displayImage(image1, "image1");
        ImageUtils.displayImage(image2, "image2");

        f1.fft();
        f2.fft();

        f1.mult(f2);

        f1.ifft();
        Image filteredImage = ImageUtils.getUnahe(ImageUtils.quadSwap(f1.getImage()));
        ImageUtils.displayImage(filteredImage, "complex plane");
    }

    public void scale(float s) {
        r.scale(s);
        g.scale(s);
        b.scale(s);
    }

    public FFT2dComplex(Image img) {
        sib = new ShortImageBean(img);
        r = new ComplexFloat1d(sib.getR());
        g = new ComplexFloat1d(sib.getG());
        b = new ComplexFloat1d(sib.getB());
    }

    public void normalize() {
        r.scale(1f / r.getLength());
        g.scale(1f / g.getLength());
        b.scale(1f / b.getLength());
    }

    public void center() {
        r.center();
        g.center();
        b.center();
    }

    /**
     * Get the real part of the complex plane in red, green and blue,
     * combine into an image.  Scale pixels to the range from
     * 0..255. If the pixels are not in that range, call <code>makeUnitary</code>
     *
     * @return java.awt.Image of real number plane
     */
    public Image getImage() {
        int w = sib.getWidth();
        int h = sib.getHeight();
        r.scale(255);
        g.scale(255);
        b.scale(255);
        sib.setR(r.getRealAs2dShort(w, h));
        sib.setG(g.getRealAs2dShort(w, h));
        sib.setB(b.getRealAs2dShort(w, h));
        return sib.getImage();
    }

    /**
     * Compute the peak values for the 3 color planes. Select
     * the max peak value and scale the color planes by 1/max.
     */
    public void makeUnitary() {
        float max = getPeakRealValue();
        float sf = 1f / max;
        r.scale(sf);
        g.scale(sf);
        b.scale(sf);
    }

    public static Image filter(Image img1, Image img2) {
        FFT2dComplex f1 = new FFT2dComplex(img1);
        FFT2dComplex f2 = new FFT2dComplex(img2);
        f1.normalize();
        f2.normalize();
        //make unitary.
        f1.scale(1f / 255f);
        f2.scale(1f / 255f);

        f1.fft();
        f2.fft();

        f1.mult(f2);
        f1.ifft();
        f1.scale(255);
        return ImageUtils.getUnahe(ImageUtils.quadSwap(f1.getImage()));
    }

    public static void testAndDisplay(Image image1, Image image2) {
        FFT2dComplex f = new FFT2dComplex(image1);
        f.fft();
        f.normalize();
        FFT2dComplex f2 = new FFT2dComplex(image2);
        ImageUtils.displayImage(image1, "image1");
        ImageUtils.displayImage(image2, "image2");
        f2.fft();
        f2.normalize();
        f.mult(f2);
        f.ifft();
        Image filteredImage = ImageUtils.quadSwap(f.getImage());
        ImageUtils.displayImage(ImageUtils.getUnahe(filteredImage), "complex plane");
    }

    public Image getPsdImage() {
        float rf[] = r.getPSD();
        float gf[] = g.getPSD();
        float bf[] = b.getPSD();
        ShortImageBean psd = new ShortImageBean(rf, gf, bf, sib.getWidth(), sib.getHeight());
        return psd.getImage();
    }

    /**
     * Multiply r,g,b in the complex frequency domain
     *
     * @param f2
     */
    public void mult(FFT2dComplex f2) {
        r.mult(f2.r);
        g.mult(f2.g);
        b.mult(f2.b);
    }

    public void ifft() {
        FFT1dComplex fftr = new FFT1dComplex();
        FFT1dComplex fftg = new FFT1dComplex();
        FFT1dComplex fftb = new FFT1dComplex();
        fftr.computeBackwardFFT(r);
        fftg.computeBackwardFFT(g);
        fftb.computeBackwardFFT(b);
    }

    public void fft() {
        FFT1dComplex fftr = new FFT1dComplex();
        FFT1dComplex fftg = new FFT1dComplex();
        FFT1dComplex fftb = new FFT1dComplex();
        fftr.computeForwardFFT(r);
        fftg.computeForwardFFT(g);
        fftb.computeForwardFFT(b);
    }


    public float getPeakRealValue() {
        float rmax = r.getMaxReal();
        float gmax = g.getMaxReal();
        float bmax = b.getMaxReal();
        float max = rmax;
        if (max < gmax) max = gmax;
        if (max < bmax) max = bmax;
        return max;
    }

    /**
     * Does the inverse of normalize
     */
    public void denormalize() {
        r.scale(r.getLength());
        g.scale(g.getLength());
        b.scale(b.getLength());
    }

    public Image getScaledImage() {
        final float sf = 255f / getPeakRealValue();
        return getScaledImage(sf);
    }

    public Image getScaledImage(float sf) {
        int w = sib.getWidth();
        int h = sib.getHeight();
        short ra[][] = sib.getR();
        short ga[][] = sib.getG();
        short ba[][] = sib.getB();
        for (int x = 0; x < w; x++)
            for (int y = 0; y < h; y++) {
                final int i = x + y * w;
                ra[x][y] = (short) (sf * r.re[i]);
                ba[x][y] = (short) (sf * b.re[i]);
                ga[x][y] = (short) (sf * g.re[i]);
            }
        sib.setR(ra);
        sib.setG(ga);
        sib.setB(ba);
        return sib.getImage();
    }

    public static Image getScaledImage(FFT2dComplex f1) {
        f1.makeUnitary();
        return f1.getImage();
    }

    public void conjugate() {
        r.conjugate();
        g.conjugate();
        b.conjugate();
    }

    // Anne MacIlvain added:
    public void flip() {
        //To change body of created methods use File | Settings | File Templates.
        int length = r.getLength();
        int flipIndex = length - 1;
        float[] rreal = new float[length];
        float[] rimag = new float[length];
        float[] greal = new float[length];
        float[] gimag = new float[length];
        float[] breal = new float[length];
        float[] bimag = new float[length];
        float[] rflipreal = new float[length];
        float[] rflipimag = new float[length];
        float[] gflipreal = new float[length];
        float[] gflipimag = new float[length];
        float[] bflipreal = new float[length];
        float[] bflipimag = new float[length];
        rreal = r.getRe();
        rimag = r.getIm();
        greal = g.getRe();
        gimag = g.getIm();
        breal = b.getRe();
        bimag = b.getIm();
        for (int ii = 0; ii < length; ii++) {
            rflipreal[flipIndex-ii] = rreal[ii];
            rflipimag[flipIndex-ii] = rimag[ii];
            gflipreal[flipIndex-ii] = greal[ii];
            gflipimag[flipIndex-ii] = gimag[ii];
            bflipreal[flipIndex-ii] = breal[ii];
            bflipimag[flipIndex-ii] = bimag[ii];
        }
        r.setRe(rflipreal);
        r.setIm(rflipimag);
        g.setRe(gflipreal);
        g.setIm(gflipimag);
        b.setRe(bflipreal);
        b.setIm(bflipimag);
    }

    // Anne MacIlvain added:
    public void removeMean() {
        //To change body of created methods use File | Settings | File Templates.
        int length = r.getLength();
        float[] rreal = new float[length];
        float[] rimag = new float[length];
        float[] greal = new float[length];
        float[] gimag = new float[length];
        float[] breal = new float[length];
        float[] bimag = new float[length];
        rreal = r.getRe();
        rimag = r.getIm();
        greal = g.getRe();
        gimag = g.getIm();
        breal = b.getRe();
        bimag = b.getIm();
        rreal[0] = (float) 0.0;
        rimag[0] = (float) 0.0;
        greal[0] = (float) 0.0;
        gimag[0] = (float) 0.0;
        breal[0] = (float) 0.0;
        bimag[0] = (float) 0.0;
        r.setRe(rreal);
        r.setIm(rimag);
        g.setRe(greal);
        g.setIm(gimag);
        b.setRe(breal);
        b.setIm(bimag);
    }

    // Anne MacIlvain added:
    public void removeDCBias() {
        //To change body of created methods use File | Settings | File Templates.
        int length = r.getLength();
        float[] rreal = new float[length];
        float[] rimag = new float[length];
        float[] greal = new float[length];
        float[] gimag = new float[length];
        float[] breal = new float[length];
        float[] bimag = new float[length];
        float dcBiasRReal = rreal[0];
        float dcBiasRImag = rimag[0];
        float dcBiasGReal = greal[0];
        float dcBiasGImag = gimag[0];
        float dcBiasBReal = breal[0];
        float dcBiasBImag = bimag[0];
        rreal = r.getRe();
        rimag = r.getIm();
        greal = g.getRe();
        gimag = g.getIm();
        breal = b.getRe();
        bimag = b.getIm();
        for (int ii = 0; ii < length; ii++) {
            rreal[ii] = rreal[ii] - dcBiasRReal;
            rimag[ii] = rimag[ii] - dcBiasRImag;
            greal[ii] = greal[ii] - dcBiasGReal;
            gimag[ii] = gimag[ii] - dcBiasGImag;
            breal[ii] = breal[ii] - dcBiasBReal;
            bimag[ii] = bimag[ii] - dcBiasBImag;
        }
        r.setRe(rreal);
        r.setIm(rimag);
        g.setRe(greal);
        g.setIm(gimag);
        b.setRe(breal);
        b.setIm(bimag);
    }
}
